-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("questsTooltip", "AceEvent-3.0")

-----------------------
--- Library loading ---
-----------------------
local LSM = LibStub("LibSharedMedia-3.0")
local QTIP = LibStub("LibQTip-1.0")
local LDB = LibStub("LibDataBroker-1.1")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local pairs, ipairs, unpack, tonumber, format, sub, tconcat = pairs, ipairs, unpack, tonumber, string.format, string.sub, table.concat
local QuestLogPushQuest, GetQuestLink, SelectQuestLogEntry, GetQuestLogSelection, WorldMap_OpenToQuest = QuestLogPushQuest, GetQuestLink, SelectQuestLogEntry, GetQuestLogSelection, WorldMap_OpenToQuest
local ItemRefTooltip, ChatEdit_InsertLink, GetItemCount, GetItemCooldown, LoadAddOn = ItemRefTooltip, ChatEdit_InsertLink, GetItemCount, GetItemCooldown, LoadAddOn
local CreateFrame, CreateFont, SetFont, ToggleFrame, ShowUIPanel = CreateFrame, CreateFont, SetFont, ToggleFrame, ShowUIPanel
local UIParent, IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown = UIParent, IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown
local GetAbandonQuestItems, GetAbandonQuestName, SetAbandonQuest = GetAbandonQuestItems, GetAbandonQuestName, SetAbandonQuest

-----------------
--- Variables ---
-----------------
local _

module.tempTable = {}
module.maxQuests = MAX_QUESTS
module.minus = "Interface\\Buttons\\UI-MinusButton-Up"
module.plus = "Interface\\Buttons\\UI-PlusButton-Up"
module.CELL_MARGIN_H = 6
module.hToFontTQ2 = CreateFont("hToFontTQ2")
module.qToFontTQ2 = CreateFont("qToFontTQ2")
module.qoToFontTQ2 = CreateFont("qoToFontTQ2")
module.interactivityAction = {
	mouseRightShift = core.LC["Shift + Right click"],
	mouseRightControl = core.LC["Ctrl + Right click"],
	mouseRightAlt = core.LC["Alt + Right click"],
	mouseRight = core.LC["Right click"],
	mouseLeftShift = core.LC["Shift + Left click"],
	mouseLeftControl = core.LC["Ctrl + Left click"],
	mouseLeftAlt = core.LC["Alt + Left click"],
	mouseLeft = core.LC["Left click"],
	mouseMiddleShift = core.LC["Shift + Middle click"],
	mouseMiddleControl = core.LC["Ctrl + Middle click"],
	mouseMiddleAlt = core.LC["Alt + Middle click"],
	mouseMiddle = core.LC["Middle click"],
}
module.interactivityList = {
	["chatLink"] = core.LC["Insert quest link in chat"],
	["objectiveProgress"] = core.LC["Insert quest objectives progress in chat"],
	["macroUpdate"] = core.LC["Update special quest item macro"],
	["questTooltip"] = core.LC["Display of quest tooltip"],
	["shareQuest"] = core.LC["Share quest"],
	["trackQuest"] = core.LC["Track quest"],
	["abandonQuest"] = core.LC["Abandon quest"],
	["displayQuest"] = core.LC["Diplay quest log"],
	["displayQuestInfo"] = core.LC["Display quest on map"],
	["nothing"] = core.LC["Do nothing"],
}

-----------------------
--- Tools functions ---
-----------------------
local function normalizeWidth(min, max)
	if max < min then return max else return min end
end

local function normalizeColorComponent(component)
	if component > 1 then return 1 elseif component < 0 then return 0 else return component end
end

local function getColorFromCompletionHEX(percent)
	if percent <= 0 then
		return format("%02x%02x%02x", 255*core.db.profile.quests.rj0, 255*core.db.profile.quests.gj0, 255*core.db.profile.quests.bj0)
	elseif percent >= 1 then
		return format("%02x%02x%02x", 255*core.db.profile.quests.rj3, 255*core.db.profile.quests.gj3, 255*core.db.profile.quests.bj3)
	elseif percent <= 0.5 then
		return format("%02x%02x%02x", 255*normalizeColorComponent(core.db.profile.quests.ri1*percent + core.db.profile.quests.rj1), 255*normalizeColorComponent(core.db.profile.quests.gi1*percent + core.db.profile.quests.gj1), 255*normalizeColorComponent(core.db.profile.quests.bi1*percent + core.db.profile.quests.bj1))
	else
		return format("%02x%02x%02x", 255*normalizeColorComponent(core.db.profile.quests.ri2*percent + core.db.profile.quests.rj2), 255*normalizeColorComponent(core.db.profile.quests.gi2*percent + core.db.profile.quests.gj2), 255*normalizeColorComponent(core.db.profile.quests.bi2*percent + core.db.profile.quests.bj2))
	end
end

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		showNumCompletedQuests = true,
		showNumQuests = true,
		showMaxQuest = true,
		colorizeQuestsCounter = true,
		hideCompletedObjective = 2,
		headerFont = LSM.DefaultMedia.font,
		headerFontSize = 12,
		headerFontOutline = "",
		questFont = LSM.DefaultMedia.font,
		questFontSize = 10,
		questFontOutline = "",
		questObjectiveFont = LSM.DefaultMedia.font,
		questObjectiveFontSize = 8,
		questObjectiveFontOutline = "",
		maxWidth = 250,
		minWidth = 250,
		statusMaxWidth = 70,
		statusMinWidth = 70,
		hasMaxHeight = true,
		maxHeight = 600,
		scale = 1,
		showQuestItemIcon = true,
		questItemIconSize = 20,
		questsCellPadding = 5,
		objectivesCellPadding = 10,
		rightCellPadding = 5,
		showCollapseButton = true,
		collapseButtonSize = 14,
		colorizeHeaderName = true,
		defaultHeaderNameColor = "ffffff",
		hideZoneLevelRange = false,
		hideQuestTag = false,
		colorizeHeaderLevel = true,
		defaultHeaderLevelColor = "ffffff",
		colorizeQuestName = true,
		defaultQuestNameColor = "ffffff",
		colorizeQuestLevel = true,
		defaultQuestLevelColor = "ffffff",
		colorizeQuestObjective = true,
		defaultQuestObjectiveColor = "ffffff",
		colorizeQuestObjectiveCounter = true,
		defaultQuestDescriptionColor = "ffffff",
		showHints = true,
		overTooltip = true,
		mouseRightShift = "chatLink",
		mouseRightControl = "objectiveProgress",
		mouseRightAlt = "macroUpdate",
		mouseRight = "questTooltip",
		mouseLeftShift = "trackQuest",
		mouseLeftControl = "shareQuest",
		mouseLeftAlt = "abandonQuest",
		mouseLeft = "displayQuest",
		mouseMiddleShift = "displayQuestInfo",
		mouseMiddleControl = "displayQuestInfo",
		mouseMiddleAlt = "displayQuestInfo",
		mouseMiddle = "displayQuestInfo",
		ldbMouseRightShift = "openOptions",
		ldbMouseRightControl = "openOptions",
		ldbMouseRightAlt = "openOptions",
		ldbMouseRight = "openOptions",
		ldbMouseLeftShift = "toggleQuestTracker",
		ldbMouseLeftControl = "toggleAnnouncementModule",
		ldbMouseLeftAlt = "toggleAchievementTracker",
		ldbMouseLeft = "toggleQuestsLog",
		ldbMouseMiddleShift = "toggleQuestTracker",
		ldbMouseMiddleControl = "toggleAnnouncementModule",
		ldbMouseMiddleAlt = "toggleAchievementTracker",
		ldbMouseMiddle = "toggleQuestsLog",
	}
}

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("questsTooltip", defaults, "Default")
	self.db.char.collapseAll = self.db.char.collapseAll or 0
	self.db.char.zoneDisplay = self.db.char.zoneDisplay or {}
	self:SetEnabledState(core:getModuleEnabled("questsTooltip"))
	core.moduleState["questsTooltip"] = core:getModuleEnabled("questsTooltip")
	LSM.RegisterCallback(self, "LibSharedMedia_Registered", "updateUsedMedia")
end

function module:OnDisable()
	core.UnregisterAllCallbacks(self)
	self:Refresh()
end

function module:Refresh()
	if not core.moduleState["questsTooltip"] then return end
	self.hToFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.headerFont), self.db.profile.headerFontSize, self.db.profile.headerFontOutline)
	self.qToFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.questFont), self.db.profile.questFontSize, self.db.profile.questFontOutline)
	self.qoToFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.questObjectiveFont), self.db.profile.questObjectiveFontSize, self.db.profile.questObjectiveFontOutline)
	self.collapseButtonSize = self.db.profile.showCollapseButton and self.db.profile.collapseButtonSize or 0
	self.maxWidth = self.db.profile.maxWidth
	self.minWidth = normalizeWidth(self.db.profile.minWidth, self.maxWidth)
	self.statusMaxWidth = self.db.profile.statusMaxWidth
	self.statusMinWidth = normalizeWidth(self.db.profile.statusMinWidth, self.statusMaxWidth)
	self.totalMaxWidth = self.maxWidth + self.statusMaxWidth + self.CELL_MARGIN_H
	self.totalMinWidth = normalizeWidth(self.minWidth + self.statusMinWidth + self.CELL_MARGIN_H, self.totalMaxWidth)
	if self.db.profile.showQuestItemIcon then
		self.questItemIconSize = self.db.profile.questItemIconSize
		self.questsCellPadding = self.db.profile.questsCellPadding + self.questItemIconSize
		self.objectivesCellPadding = self.db.profile.objectivesCellPadding + self.questItemIconSize
	else
		self.questItemIconSize = 0
		self.questsCellPadding = self.db.profile.questsCellPadding
		self.objectivesCellPadding = self.db.profile.objectivesCellPadding
	end

end

function module:updateUsedMedia(event, mediatype, key)
	if not core.moduleState["questsTooltip"] then return end
	if mediatype == "font" then
		local updateTooltip = nil
		if key == self.db.profile.headerFont then
			updateTooltip = true
			self.hToFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.headerFont), self.db.profile.headerFontSize, self.db.profile.headerFontOutline)
		end
		if key == self.db.profile.questFont then
			updateTooltip = true
			self.qToFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.questFont), self.db.profile.questFontSize, self.db.profile.questFontOutline)
		end
		if key == self.db.profile.questObjectiveFont then
			updateTooltip = true
			self.qoToFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.questObjectiveFont), self.db.profile.questObjectiveFontSize, self.db.profile.questObjectiveFontOutline)
		end
		if self.tooltipIsShown and updateTooltip then
			self:updateTooltip()
		end
	end
end

local function hideTooltip()
	if not module.tip or module.tip:IsMouseOver() then return end
	module.tooltipIsShown = nil
	module.tip:SetScript("OnUpdate", nil)
	QTIP:Release(module.tip)
	module.tip = nil
end

---------------------------
--- Main Module Program ---
---------------------------
function module:questLogUpdated(event, ldbText, tooltip)
	if not core.moduleState["questsTooltip"] then return end
	if ldbText then
		local tIndex = 0
		self.ldbColor = nil
		if self.db.profile.showNumCompletedQuests then
			tIndex = tIndex + 1
			self.tempTable[tIndex] = core.numCompletedQuests
			if core.numQuests > 0 then
				self.ldbColor = getColorFromCompletionHEX(core.numCompletedQuests/core.numQuests)
			else
				self.ldbColor = "ffffff"
			end
		end
		if self.db.profile.showNumQuests then
			tIndex = tIndex + 1
			self.tempTable[tIndex] = core.numQuests
			self.ldbColor = self.ldbColor or getColorFromCompletionHEX(core.numQuests/self.maxQuests)
		end
		if self.db.profile.showMaxQuest then
			tIndex = tIndex + 1
			self.tempTable[tIndex] = self.maxQuests
			self.ldbColor = self.ldbColor or "ffffff"
		end
		if tIndex > 0 then
			if self.db.profile.colorizeQuestsCounter then
				self.dataobj.text = "|cff"..self.ldbColor..tconcat(self.tempTable, "/", 1, tIndex).."|r"
			else
				self.dataobj.text = tconcat(self.tempTable, "/", 1, tIndex)
			end
		else
			self.dataobj.text = ""
		end
	end
	if tooltip and self.tooltipIsShown then
		self:updateTooltip()
	end
end

---------------------------------------------------
--- QTIP Mouse Handler and Custom Cell Provider ---
---------------------------------------------------
local interactivity = {
	chatLink = function(cValue1)
		if core.quests[cValue1].complete[core.playerName] and core.quests[cValue1].id then
			ChatFrameEditBox:Show()
			ChatEdit_InsertLink(GetQuestLink(core.quests[cValue1].id))
		end
	end,
	objectiveProgress = function(cValue1)
		if core.quests[cValue1].numObjectives > 0 and core.quests[cValue1].complete[core.playerName] then
			ChatFrameEditBox:Show()
			for objective, qoData in pairs(core.questsObjectives[cValue1]) do
				ChatFrameEditBox:Insert("{"..qoData.desc.." "..qoData.got[core.playerName].."/"..qoData.need.."}")
			end
		end
	end,
	macroUpdate = function(cValue1)
		core:generateSpecialItemMacro(cValue1)
	end,
	questTooltip = function(cValue1)
		ItemRefTooltip:ClearLines()
		ShowUIPanel(ItemRefTooltip)
		if not ItemRefTooltip:IsVisible() then
			ItemRefTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
		end
		ItemRefTooltip:SetHyperlink("quest:"..cValue1)
	end,
	shareQuest = function(cValue1)
		if core.numPartyMembers > 0 and core.quests[cValue1].complete[core.playerName] and core.quests[cValue1].id and core.quests[cValue1].sharable then
			local wasSelected = GetQuestLogSelection()
			SelectQuestLogEntry(core.quests[cValue1].id)
			QuestLogPushQuest()
			SelectQuestLogEntry(wasSelected)
		end
	end,
	trackQuest = function(cValue1)
		core.cbh:Fire("toggleTrackedQuest", cValue1)
		module:updateTooltip()
	end,
	abandonQuest = function(cValue1)
		if core.quests[cValue1].complete[core.playerName] and core.quests[cValue1].id then
			local wasSelected = GetQuestLogSelection()
			SelectQuestLogEntry(core.quests[cValue1].id)
			SetAbandonQuest()
			local items = GetAbandonQuestItems()
			if items then
				StaticPopup_Hide("ABANDON_QUEST")
				StaticPopup_Show("ABANDON_QUEST_WITH_ITEMS", GetAbandonQuestName(), items)
			else
				StaticPopup_Hide("ABANDON_QUEST_WITH_ITEMS")
				StaticPopup_Show("ABANDON_QUEST", GetAbandonQuestName())
			end
			SelectQuestLogEntry(wasSelected)
		end
	end,
	displayQuest = function(cValue1)
		core:displayQuestInformations(cValue1)
	end,
	displayQuestInfo = function(cValue1)
		WorldMap_OpenToQuest(cValue1)
	end,
	nothing = function(cValue1)
		return
	end
}

local questOnMouseDown = function(cell, button)
	local cType1, cValue1 = cell.cType1, cell.cValue1
	if cType1 == "header" then
		if module.db.char.collapseAll == 1 then
			module.db.char.collapseAll = 0
			for zone in pairs(module.db.char.zoneDisplay) do
				module.db.char.zoneDisplay[zone] = 0
			end
		else
			module.db.char.collapseAll = 1
			for zone in pairs(module.db.char.zoneDisplay) do
				module.db.char.zoneDisplay[zone] = 1
			end
		end
		module:updateTooltip()
	elseif cType1 == "zone" then
		if module.db.char.zoneDisplay[cValue1] == 1 then
			module.db.char.zoneDisplay[cValue1] = 0
			module.db.char.collapseAll = 0
		else
			module.db.char.zoneDisplay[cValue1] = 1
			module.db.char.collapseAll = 1
		end
		module:updateTooltip()
	elseif cType1 == "quest" then
		if button == "RightButton" then
			if IsShiftKeyDown() then
				interactivity[module.db.profile.mouseRightShift](cValue1)
			elseif IsControlKeyDown() then
				interactivity[module.db.profile.mouseRightControl](cValue1)
			elseif IsAltKeyDown() then
				interactivity[module.db.profile.mouseRightAlt](cValue1)
			else
				interactivity[module.db.profile.mouseRight](cValue1)
			end
		elseif button == "LeftButton" then
			if IsShiftKeyDown() then
				interactivity[module.db.profile.mouseLeftShift](cValue1)
			elseif IsControlKeyDown() then
				interactivity[module.db.profile.mouseLeftControl](cValue1)
			elseif IsAltKeyDown() then
				interactivity[module.db.profile.mouseLeftAlt](cValue1)
			else
				interactivity[module.db.profile.mouseLeft](cValue1)
			end
		elseif button == "MiddleButton" then
			if IsShiftKeyDown() then
				interactivity[module.db.profile.mouseMiddleShift](cValue1)
			elseif IsControlKeyDown() then
				interactivity[module.db.profile.mouseMiddleControl](cValue1)
			elseif IsAltKeyDown() then
				interactivity[module.db.profile.mouseMiddleAlt](cValue1)
			else
				interactivity[module.db.profile.mouseMiddle](cValue1)
			end
		end
	end
end

local function questOnEnter(cell)
	if module.db.profile.overTooltip and cell.cType1 == "quest" then
		GameTooltip:SetOwner(module.tip, module.tipAnchor)
		GameTooltip:SetHyperlink("quest:"..cell.cValue1)
	end
end

local function questOnLeave(cell)
	GameTooltip:Hide()
end

local function buttonOnEnter(self)
	local link = self.link
	if not link then return end
	GameTooltip:SetOwner(self, "ANCHOR_LEFT")
	GameTooltip:SetHyperlink(link)
end

local function buttonOnLeave()
	GameTooltip:Hide()
end

local function buttonOnClick(button)
	local category = button.category
	if module.db.char.zoneDisplay[category] == 1 then
		module.db.char.zoneDisplay[category] = 0
	else
		module.db.char.zoneDisplay[category] = 1
	end
	module:updateTooltip()
end

local function headerOnClick(button)
	if module.db.char.collapseAll == 1 then
		module.db.char.collapseAll = 0
		for zone in pairs(module.db.char.zoneDisplay) do
			module.db.char.zoneDisplay[zone] = 0
		end
	else
		module.db.char.collapseAll = 1
		for zone in pairs(module.db.char.zoneDisplay) do
			module.db.char.zoneDisplay[zone] = 1
		end
	end
	module:updateTooltip()
end

local function wrongButtonOnClick(button)
	core:Print(core.LC["This button is not clickable, you can use tomQuest2 macro system instead"])
end

local function itemOnEvent(self, event, ...)
	if event == "BAG_UPDATE_COOLDOWN" then
		local start, duration, enable = GetItemCooldown(self.link)
		if start > 0 and duration > 0 and enable > 0 then
			self.coolDown:SetCooldown(start, duration)
			self.coolDown:Show()
		else
			self.coolDown:Hide()
		end
	end
end

local cProvider, cPrototype, baseCellPrototype = QTIP:CreateCellProvider(QTIP.LabelProvider)
function cPrototype:SetupCell(tooltip, value, justification, font, lPad, rPad, maxWidth, minWidth, cType1, cType2, cValue1, cValue2)
	local width, height = baseCellPrototype.SetupCell(self, tooltip, value, justification, font, lPad, rPad + 10, maxWidth, minWidth)
	if self.button then
		core:recycleButton(self.button)
		self.button = nil
	end
	self:EnableMouse(true)
	self.cType1, self.cValue1 = cType1, cValue1
	self:SetScript("OnEnter", questOnEnter)
	self:SetScript("OnLeave", questOnLeave)
	self:SetScript("OnMouseDown", questOnMouseDown)
	if cType2 == "icon" then
		if cType1 == "quest" then
			if cValue2 ~= 0 and module.db.profile.showQuestItemIcon and (module.db.profile.hideCompletedObjective == 3 or (module.db.profile.hideCompletedObjective == 2 and core.quests[cValue1].partyComplete == 0) or (module.db.profile.hideCompletedObjective == 1 and core.quests[cValue1].playerComplete == 0)) then
				if core.quests[cValue1].item then
					self.button = core:getItemButton()
					self.button:SetID(core.quests[cValue1].id)
					self.button:SetScript("OnEvent", itemOnEvent)
					local itemCount = GetItemCount(cValue2)
					if itemCount > 1 then
						self.button.count:SetText(itemCount)
						self.button.count:Show()
					end
					self.button:RegisterEvent("BAG_UPDATE_COOLDOWN")
					local start, duration, enable = GetItemCooldown(cValue2)
					if start > 0 and duration > 0 and enable > 0 then
						self.button.coolDown:SetCooldown(start, duration)
						self.button.coolDown:Show()
					else
						self.button.coolDown:Hide()
					end
				else
					self.button = core:getNormalButton()
					self.button:SetScript("OnEnter", buttonOnEnter)
					self.button:SetScript("OnLeave", buttonOnLeave)
					self.button:SetScript("OnClick", wrongButtonOnClick)
				end
				self.button.iconTexture:SetTexture(core.quests[cValue1].specialItemIcon)
				self.button.link = cValue2
				self.button:SetScale(module.db.profile.questItemIconSize/self.button.height)
				self.button:SetParent(self)
				self.button:SetPoint("TOPLEFT", self, "TOPLEFT", module.db.profile.questsCellPadding, 0)
				self.button:SetToplevel(true)
				self.button:Show()
			end
			if core.quests[cValue1].tracked then
				self.check = core:getCollapseButton()
				self.check.iconTexture:SetTexture("Interface\\Buttons\\UI-CheckBox-Check")
				self.check:SetHeight(height)
				self.check:SetWidth(height)
				self.check:SetParent(self)
				self.check:SetPoint("TOPLEFT", self, "TOPLEFT", self.fontString:GetStringWidth() + module.questsCellPadding, 0)
				self.check:SetToplevel(true)
				self.check:Show()
			end
		elseif cType1 == "zone" then
			if module.db.profile.showCollapseButton then
				self.button = core:getCollapseButton()
				self.button:RegisterForClicks("AnyDown")
				self.button.iconTexture:SetTexture(cValue2)
				self.button.category = cValue1
				self.button:SetScript("OnClick", buttonOnClick)
				if module.db.profile.collapseButtonSize > height then
					self.button:SetHeight(height)
					self.button:SetWidth(height)
				else
					self.button:SetHeight(module.db.profile.collapseButtonSize)
					self.button:SetWidth(module.db.profile.collapseButtonSize)
				end
				self.button:SetParent(self)
				self.button:SetPoint("TOPLEFT", self, "TOPLEFT")
				self.button:SetToplevel(true)
				self.button:Show()
			end
		elseif cType1 == "header" then
			if module.db.profile.showCollapseButton then
				self.button = core:getCollapseButton()
				self.button:RegisterForClicks("AnyDown")
				self.button.iconTexture:SetTexture(cValue2)
				self.button:SetScript("OnClick", headerOnClick)
				if module.db.profile.collapseButtonSize > height then
					self.button:SetHeight(height)
					self.button:SetWidth(height)
				else
					self.button:SetHeight(module.db.profile.collapseButtonSize)
					self.button:SetWidth(module.db.profile.collapseButtonSize)
				end
				self.button:SetParent(self)
				self.button:SetPoint("TOPRIGHT", self, "TOPLEFT", (width - self.fontString:GetStringWidth())/2 - 10, 0)
				self.button:SetToplevel(true)
				self.button:Show()
			end
		end
	end
	return width, height
end

function cPrototype:ReleaseCell()
	if self.button then
		core:recycleButton(self.button)
		self.button = nil
	end
	if self.check then
		core:recycleButton(self.check)
		self.check = nil
	end
	self:EnableMouse(false)
	self.cType1, self.cValue1 = nil, nil
	self:SetScript("OnEnter", nil)
	self:SetScript("OnLeave", nil)
	self:SetScript("OnMouseDown", nil)
end

function module:updateTooltip()
	self.tip = QTIP:Acquire("tomQuest2ToolTip", 2, "LEFT", "RIGHT")
	self.tip:Clear()
	self.tooltipIsShown = true
	self.tip:SetScale(self.db.profile.scale)
	local tempLine = self.tip:AddLine()
	self.tip:SetCell(tempLine, 1, "tomQuest2", self.hToFontTQ2, "CENTER", 2, cProvider, 0, 0, self.totalMaxWidth, self.totalMinWidth, "header", "icon", nil, self.db.char.collapseAll == 0 and self.plus or self.minus)
	self.tip:AddLine(" ")
	for _, zone in ipairs(core.sortedZones) do
		self.db.char.zoneDisplay[zone] = self.db.char.zoneDisplay[zone] or 1
		tempLine = self.tip:AddLine()
		if not self.db.profile.hideZoneLevelRange then
			self.tip:SetCell(tempLine, 1, "|cff"..(self.db.profile.colorizeHeaderName and core.zones[zone].color or self.db.profile.defaultHeaderNameColor)..zone.."|r |cff"..(self.db.profile.colorizeHeaderLevel and core.zones[zone].color or self.db.profile.defaultHeaderLevelColor)..core.zones[zone].level.."|r", self.hToFontTQ2, "LEFT", 2, cProvider, self.collapseButtonSize, self.db.profile.rightCellPadding, self.totalMaxWidth, self.totalMinWidth, "zone", "icon", zone, self.db.char.zoneDisplay[zone] == 1 and self.minus or self.plus)
		else
			self.tip:SetCell(tempLine, 1, "|cff"..(self.db.profile.colorizeHeaderName and core.zones[zone].color or self.db.profile.defaultHeaderNameColor)..zone.."|r", self.hToFontTQ2, "LEFT", 2, cProvider, self.collapseButtonSize, self.db.profile.rightCellPadding, self.totalMaxWidth, self.totalMinWidth, "zone", "icon", zone, self.db.char.zoneDisplay[zone] == 1 and self.minus or self.plus)
		end
		if self.db.char.zoneDisplay[zone] == 1 then
			for _, uid in ipairs(core.sortedQuestsByZone[zone]) do
				tempLine = self.tip:AddLine()
				if core.quests[uid].playersOnQuest > 1 then
					if self.db.profile.hideQuestTag then
						self.tip:SetCell(tempLine, 1, "|cff"..(self.db.profile.colorizeQuestName and core.quests[uid].color or self.db.profile.defaultQuestNameColor)..core.quests[uid].title.."|r <"..core.quests[uid].playersOnQuest..">", self.qToFontTQ2, "LEFT", 1, cProvider, self.questsCellPadding, 0, self.maxWidth, self.minWidth, "quest", "icon", uid, core.quests[uid].specialItemLink or 0)
					else
						self.tip:SetCell(tempLine, 1, "|cff"..(self.db.profile.colorizeQuestLevel and core.quests[uid].color or self.db.profile.defaultQuestLevelColor)..core.quests[uid].tag.."|r |cff"..(self.db.profile.colorizeQuestName and core.quests[uid].color or self.db.profile.defaultQuestNameColor)..core.quests[uid].title.."|r <"..core.quests[uid].playersOnQuest..">", self.qToFontTQ2, "LEFT", 1, cProvider, self.questsCellPadding, 0, self.maxWidth, self.minWidth, "quest", "icon", uid, core.quests[uid].specialItemLink or 0)
					end
				else
					if self.db.profile.hideQuestTag then
						self.tip:SetCell(tempLine, 1, "|cff"..(self.db.profile.colorizeQuestName and core.quests[uid].color or self.db.profile.defaultQuestNameColor)..core.quests[uid].title.."|r", self.qToFontTQ2, "LEFT", 1, cProvider, self.questsCellPadding, 0, self.maxWidth, self.minWidth, "quest", "icon", uid, core.quests[uid].specialItemLink or 0)
					else
						self.tip:SetCell(tempLine, 1, "|cff"..(self.db.profile.colorizeQuestLevel and core.quests[uid].color or self.db.profile.defaultQuestLevelColor)..core.quests[uid].tag.."|r |cff"..(self.db.profile.colorizeQuestName and core.quests[uid].color or self.db.profile.defaultQuestNameColor)..core.quests[uid].title.."|r", self.qToFontTQ2, "LEFT", 1, cProvider, self.questsCellPadding, 0, self.maxWidth, self.minWidth, "quest", "icon", uid, core.quests[uid].specialItemLink or 0)
					end
				end
				self.tip:SetCell(tempLine, 2, core.quests[uid].pqStatus, self.qToFontTQ2, "RIGHT", 1, cProvider, 0, self.db.profile.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "quest", nil, uid)
				if core.questsObjectives[uid] then
					for objective, qoData in ipairs(core.questsObjectives[uid]) do
						if self.db.profile.hideCompletedObjective == 3 or (self.db.profile.hideCompletedObjective == 2 and qoData.partyComplete == 0) or (self.db.profile.hideCompletedObjective == 1 and qoData.playerComplete == 0) then
							tempLine = self.tip:AddLine()
							if self.db.profile.colorizeQuestObjective then
								self.tip:SetCell(tempLine, 1, "|cff"..qoData.qoColor..qoData.desc.."|r", self.qoToFontTQ2, "LEFT", 1, cProvider, self.objectivesCellPadding, 0, self.maxWidth, self.minWidth, "quest", nil, uid)
							else
								self.tip:SetCell(tempLine, 1, "|cff"..self.db.profile.defaultQuestObjectiveColor..qoData.desc.."|r", self.qoToFontTQ2, "LEFT", 1, cProvider, self.objectivesCellPadding, 0, self.maxWidth, self.minWidth, "quest", nil, uid)
							end
							if self.db.profile.colorizeQuestObjectiveCounter then
								self.tip:SetCell(tempLine, 2, qoData.qoStatus, self.qoToFontTQ2, "RIGHT", 1, cProvider, 0, self.db.profile.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "quest", nil, uid)
							else
								self.tip:SetCell(tempLine, 2, qoData.qoStatusNC, self.qoToFontTQ2, "RIGHT", 1, cProvider, 0, self.db.profile.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "quest", nil, uid)
							end
						end
					end
				elseif core.quests[uid].objective then
					tempLine = self.tip:AddLine()
					self.tip:SetCell(tempLine, 1, "|cff"..self.db.profile.defaultQuestDescriptionColor..core.quests[uid].objective.."|r", self.qoToFontTQ2, "LEFT", 2, cProvider, self.objectivesCellPadding, self.db.profile.rightCellPadding, self.totalMaxWidth, self.totalMinWidth, "quest", nil, uid)
				end
			end
		end
	end
	self.tip:AddLine(" ")
	if self.db.profile.showHints then
		for k, v in pairs(self.interactivityAction) do
			local action = self.db.profile[k]
			if action ~= "nothing" then
				tempLine = self.tip:AddLine()
				self.tip:SetCell(tempLine, 1, "|cffeda55f"..v.."|r |cff19ff19"..self.interactivityList[action]..".|r", self.qoToFontTQ2, "LEFT", 2, cProvider, 0, self.db.profile.rightCellPadding, self.totalMaxWidth, self.totalMinWidth)
			end
		end
	end
	self.tip:SmartAnchorTo(self.anchor)
	self.tip:UpdateScrolling(self.db.profile.hasMaxHeight and self.db.profile.maxHeight or nil)
	self.tip:Show()
end

local function dataobjOnEnter(frame)
	if not module.tooltipIsShown then
		module.anchor = frame
		module.tipAnchor = frame:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
		module:updateTooltip()
	end
end

local function dataobjOnLeave(frame)
	if module.tip and module.tip:IsMouseOver() then
		module.tip:SetScript("OnUpdate", hideTooltip)
	else
		hideTooltip()
	end
end

local ldbOnClick = {
	openOptions = function()
		LoadAddOn("tomQuest2_Config")
		InterfaceOptionsFrame_OpenToCategory(core.profilesOptionFrame)
		InterfaceOptionsFrame_OpenToCategory(core.mainOptionFrame)
	end,
	toggleQuestTracker = function()
		tQ2toggleQuestTracker()
	end,
	toggleAchievementTracker = function()
		tQ2toggleAchievementTracker()
	end,
	toggleAnnouncementModule = function()
		core:toggleModule("announce")
	end,
	toggleQuestsLog = function()
		ToggleFrame(QuestLogFrame)
	end,
	toggleSyncPartyModule = function()
		core:toggleModule("syncParty")
	end,
	toggleQuestsTrackerModule = function()
		core:toggleModule("questsTracker")
	end,
	toggleAchievementTrackerModule = function()
		core:toggleModule("achievementTracker")
	end,
	nothing = function(value)
		return
	end,
}

local function dataobjOnClick(frame, button)
	if button == "RightButton" then
		if IsShiftKeyDown() then
			ldbOnClick[module.db.profile.ldbMouseRightShift]()
		elseif IsControlKeyDown() then
			ldbOnClick[module.db.profile.ldbMouseRightControl]()
		elseif IsAltKeyDown() then
			ldbOnClick[module.db.profile.ldbMouseRightAlt]()
		else
			ldbOnClick[module.db.profile.ldbMouseRight]()
		end
	elseif button == "LeftButton" then
		if IsShiftKeyDown() then
			ldbOnClick[module.db.profile.ldbMouseLeftShift]()
		elseif IsControlKeyDown() then
			ldbOnClick[module.db.profile.ldbMouseLeftControl]()
		elseif IsAltKeyDown() then
			ldbOnClick[module.db.profile.ldbMouseLeftAlt]()
		else
			ldbOnClick[module.db.profile.ldbMouseLeft]()
		end
	elseif button == "MiddleButton" then
		if IsShiftKeyDown() then
			ldbOnClick[module.db.profile.ldbMouseMiddleShift]()
		elseif IsControlKeyDown() then
			ldbOnClick[module.db.profile.ldbMouseMiddleControl]()
		elseif IsAltKeyDown() then
			ldbOnClick[module.db.profile.ldbMouseMiddleAlt]()
		else
			ldbOnClick[module.db.profile.ldbMouseMiddle]()
		end
	end
end

function module:OnEnable()
	if not self.dataobj then
		self.dataobj = LDB:NewDataObject("tomQuest2_QuestsTooltip", {type = "data source", icon = "Interface\\AddOns\\tomQuest2\\media\\icon"})
		self.dataobj.OnEnter = dataobjOnEnter
		self.dataobj.OnLeave = dataobjOnLeave
		self.dataobj.OnClick = dataobjOnClick
	end
	core.RegisterCallback(self, "questLogUpdated")
	self.hToFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.headerFont), self.db.profile.headerFontSize, self.db.profile.headerFontOutline)
	self.qToFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.questFont), self.db.profile.questFontSize, self.db.profile.questFontOutline)
	self.qoToFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.questObjectiveFont), self.db.profile.questObjectiveFontSize, self.db.profile.questObjectiveFontOutline)
	self.collapseButtonSize = self.db.profile.showCollapseButton and self.db.profile.collapseButtonSize or 0
	self.maxWidth = self.db.profile.maxWidth
	self.minWidth = normalizeWidth(self.db.profile.minWidth, self.maxWidth)
	self.statusMaxWidth = self.db.profile.statusMaxWidth
	self.statusMinWidth = normalizeWidth(self.db.profile.statusMinWidth, self.statusMaxWidth)
	self.totalMaxWidth = self.maxWidth + self.statusMaxWidth + self.CELL_MARGIN_H
	self.totalMinWidth = normalizeWidth(self.minWidth + self.statusMinWidth + self.CELL_MARGIN_H, self.totalMaxWidth)
	if self.db.profile.showQuestItemIcon then
		self.questItemIconSize = self.db.profile.questItemIconSize
		self.questsCellPadding = self.db.profile.questsCellPadding + self.questItemIconSize
		self.objectivesCellPadding = self.db.profile.objectivesCellPadding + self.questItemIconSize
	else
		self.questItemIconSize = 0
		self.questsCellPadding = self.db.profile.questsCellPadding
		self.objectivesCellPadding = self.db.profile.objectivesCellPadding
	end
end